class PlatformNotSupported(Exception):
  pass


class SmartCardError(Exception):
  pass


class SmartCardFindKeyObjectError(SmartCardError):
  def __init__(self, key_id):
    key_id = key_id[0] if isinstance(key_id, tuple) else key_id
    super().__init__('Could not get key for key id: {}.'.format(key_id))


class SmartCardWrongPinError(SmartCardError):
  pass


class SmartCardNotPresentError(SmartCardError):
  pass


class SmartCardSigningError(SmartCardError):
  def __init__(self, data):
    super().__init__('Unable to create generate for data:\n{}\n'.format(data))



import pytest

from .pkcs11 import PKCS11
from .settings import MOCK_PYKCS11


def pytest_configure(config):
  """Register markers, ..."""
  config.addinivalue_line('markers', 'skip_smartcard')


@pytest.fixture
def pkcs11(request):
  """Return real or mocked pkcs11 library.
    When testing real library, smart card should be inserted in local PC and
    constants in `settings.py` should be changed appropriately.
  """
  if MOCK_PYKCS11:
    return PKCS11(**getattr(request, 'param', dict()))
  else:
    return None  # real PyKCS11Lib will be instantiated - test locally with sc


@pytest.fixture(autouse=True)
def skip_for_smartcard(request, pkcs11):
  """Marker to skip test if testing with real pkcs11 libraary
  """
  if request.node.get_closest_marker('skip_smartcard'):
    if not isinstance(pkcs11, PKCS11):
      pytest.skip('Skipped while testing real smart card.')



from pathlib import Path

import click

from .api import (sc_export_pub_key_pem, sc_export_x509_pem, sc_is_present,
                  sc_session, sc_sign_rsa_pkcs_pss_sha256)
from .exceptions import SmartCardError


@click.group()
def oll_sc():
  """oll-sc tool CLI"""


@oll_sc.command()
@click.argument('key_id', type=int)
@click.argument('pin')
@click.option('--output-path', '-o', type=click.Path(), default=None,
              help='The output file path to write public key pem to.')
def public_key(key_id, pin, output_path=None):
  """Extract public key from smart card in PEM format."""
  try:
    pub_key_pem_bytes = sc_export_pub_key_pem((key_id,), pin)
    pub_key_pem = pub_key_pem_bytes.decode('utf-8')

    if output_path:
      with open(output_path, 'w') as out:
        out.write(pub_key_pem)
    else:
      click.echo(pub_key_pem)

  except SmartCardError as e:
    click.echo(e)


@oll_sc.command()
def inserted():
  """Check if smart card is inserted."""
  if sc_is_present():
    click.echo('Smart card is inserted.')
  else:
    click.echo('Smart card is not inserted.')


@oll_sc.command()
@click.argument('pin')
def check_pin(pin):
  """Check smart card PIN."""
  try:
    with sc_session(pin):
      pass
    click.echo('PIN OK.')
  except SmartCardError as e:
    click.echo(e)


@oll_sc.command()
@click.argument('key_id', type=int)
@click.argument('pin')
@click.option('--input-path', '-i', type=click.Path(), default=None,
              help='Path of a file to sign.')
@click.option('--input-data', '-d', type=str, default=None,
              help='Data to sign.')
@click.option('--output-path', '-o', type=click.Path(), default=None,
              help='Path of a file to write signature to.')
def sign_rsa_pkcs_pss_sha256(key_id, pin, input_path, input_data, output_path):
  """Sign input using SHA256_RSA_PKCS_PSS mechanism."""
  # Input path overrides input data
  if input_path is not None:
    input_path = Path(input_path)
    if input_path.is_file():
      input_data = input_path.read_bytes()

  if input_data is None:
    click.echo('\nError: Missing option "--input-data" or "--input-path".')
    return

  try:
    signature = sc_sign_rsa_pkcs_pss_sha256(input_data, (key_id,), pin)

    if output_path:
      with open(output_path, 'wb') as out:
        out.write(signature)
    else:
      click.echo(signature)

  except SmartCardError as e:
    click.echo(e)


@oll_sc.command()
@click.argument('key_id', type=int)
@click.argument('pin')
@click.option('--output-path', '-o', type=click.Path(), default=None,
              help='The output file path to write public key pem to.')
def x509(key_id, pin, output_path=None):
  """Extract x509 certificate from smart card in PEM format."""
  try:
    x509_cert_bytes = sc_export_x509_pem((key_id,), pin)
    x509_cert = x509_cert_bytes.decode('utf-8')

    if output_path:
      with open(output_path, 'w') as out:
        out.write(x509_cert)
    else:
      click.echo(x509_cert)

  except SmartCardError as e:
    click.echo(e)



# Fake pkcs11 classes for simulation
import pickle
from pathlib import Path

from PyKCS11 import (CKO_CERTIFICATE, CKO_PRIVATE_KEY, CKO_PUBLIC_KEY,
                     PyKCS11Error)

from .settings import VALID_KEY_ID, VALID_MECH, VALID_PIN


def _is_valid_mechanism(mechanism):
  return mechanism._mech.mechanism == VALID_MECH._mech.mechanism and \
      mechanism._param.hashAlg == VALID_MECH._param.hashAlg and \
      mechanism._param.mgf == VALID_MECH._param.mgf and \
      mechanism._param.sLen == VALID_MECH._param.sLen


class _Session:
  """Fake pkcs11 session implementation used to test API.
  For more info:
    https://github.com/LudovicRousseau/PyKCS11/blob/master/PyKCS11/__init__.py#L851
  """

  def __init__(self, able_to_login=True):
    self._able_to_login = able_to_login
    self.logged_in = False
    self.session_closed = False

  def closeSession(self):
    self.session_closed = True

  def findObjects(self, *args):
    # If pin is wrong, return empty list
    if args[0][0][1] != VALID_KEY_ID:
      return []

    return {
        # Certificate
        CKO_CERTIFICATE: ['cert'],
        CKO_PUBLIC_KEY: ['pub_key'],
        CKO_PRIVATE_KEY: ['priv_key']
    }.get(args[0][1][1], [])

  def getAttributeValue(self, obj, *args):
    if obj == 'pub_key':
      with open(str(Path(__file__).parent / 'keys/public_key.cer'), 'rb') as der:
        return [pickle.loads(der.read())]
    elif obj == 'cert':
      with open(str(Path(__file__).parent / 'keys/x509_cert.cer'), 'rb') as der:
        return [pickle.loads(der.read())]
    else:
      return []

  def login(self, pin, user_type=None):
    if not self._able_to_login or pin != VALID_PIN:
      raise PyKCS11Error('Could not login.')
    self.logged_in = True

  def logout(self):
    if not self.logged_in:
      raise PyKCS11Error('Could not logout.')
    self.logged_in = False

  def sign(self, pk, data, mechanism):
    if not _is_valid_mechanism(mechanism):
      raise PyKCS11Error('Mechanism is not valid.')
    if not isinstance(data, bytes):
      raise TypeError()

    return b'signature'


class PKCS11:
  """Fake pkcs11 lib implementation used to test API.
  For more info:
    https://github.com/LudovicRousseau/PyKCS11/blob/master/PyKCS11/__init__.py#L453
  """

  def __init__(self, sc_inserted=True, able_to_open_session=True,
               _able_to_login=True):
    self._able_to_login = _able_to_login
    self._able_to_open_session = able_to_open_session
    self._sc_inserted = sc_inserted

  def getSlotList(self, tokenPresent=False):
    if self._sc_inserted:
      return [0]
    else:
      return []

  def openSession(self, slot, flags=0):
    if not self._able_to_open_session:
      raise PyKCS11Error('Could not open a session.')

    return _Session(self._able_to_login)



from pathlib import Path

import pytest

from oll_sc.api import (sc_export_pub_key_pem, sc_export_x509_pem,
                        sc_is_present, sc_session, sc_sign_rsa,
                        sc_sign_rsa_pkcs_pss_sha256)
from oll_sc.exceptions import (SmartCardFindKeyObjectError,
                               SmartCardNotPresentError, SmartCardSigningError,
                               SmartCardWrongPinError)

from .pkcs11 import PKCS11
from .settings import (VALID_KEY_ID, VALID_MECH, VALID_PIN, WRONG_KEY_ID,
                       WRONG_MECH, WRONG_PIN)


def test_sc_export_pub_key_valid_key_id_should_return_pem(pkcs11):
  with open(str(Path(__file__).parent / 'keys/public_key.pem'), 'rb') as pem:
    pub_key_pem = sc_export_pub_key_pem(VALID_KEY_ID, VALID_PIN, pkcs11=pkcs11)
    assert pub_key_pem == pem.read()


def test_sc_export_pub_key_wrong_key_id_should_raise_error(pkcs11):
  with pytest.raises(SmartCardFindKeyObjectError):
    sc_export_pub_key_pem(WRONG_KEY_ID, VALID_PIN, pkcs11=pkcs11)


def test_sc_export_x509_valid_key_id_should_return_cert(pkcs11):
  with open(str(Path(__file__).parent / 'keys/x509_cert.pem'), 'rb') as pem:
    x509_cert = sc_export_x509_pem(VALID_KEY_ID, VALID_PIN, pkcs11=pkcs11)
    assert x509_cert == pem.read()


def test_sc_export_x509_valid_key_id_should_should_raise_error(pkcs11):
  with pytest.raises(SmartCardFindKeyObjectError):
    sc_export_x509_pem(WRONG_KEY_ID, VALID_PIN, pkcs11=pkcs11)


def test_sc_is_present_should_return_true(pkcs11):
  assert sc_is_present(pkcs11=pkcs11)


@pytest.mark.skip_smartcard
@pytest.mark.parametrize('pkcs11', [dict(sc_inserted=False)], indirect=True)
def test_sc_is_present_should_return_false(pkcs11):
  assert not sc_is_present(pkcs11=pkcs11)


def test_sc_session_with_valid_pin_should_return_session_obj(pkcs11):
  with sc_session(VALID_PIN, pkcs11=pkcs11) as session:
    assert session

    if isinstance(pkcs11, PKCS11):
      assert session.logged_in
      assert not session.session_closed

  if isinstance(pkcs11, PKCS11):
    assert session.session_closed


@pytest.mark.skip_smartcard
@pytest.mark.parametrize('pkcs11', [dict(sc_inserted=False)], indirect=True)
def test_sc_session_without_sc_inserted_should_raise_error(pkcs11):
  with pytest.raises(SmartCardNotPresentError):
    with sc_session(VALID_PIN, pkcs11=pkcs11):
      pass


def test_sc_session_wrong_pin_should_raise_error(pkcs11):
  with pytest.raises(SmartCardWrongPinError):
    with sc_session(WRONG_PIN, pkcs11=pkcs11):
      pass


def test_sc_sign_rsa_wrong_key_id_should_raise_error(pkcs11):
  with pytest.raises(SmartCardFindKeyObjectError):
    sc_sign_rsa('test', VALID_MECH, WRONG_KEY_ID, VALID_PIN, pkcs11=pkcs11)


def test_sc_sign_rsa_wrong_mechanism_should_raise_error(pkcs11):
  with pytest.raises(SmartCardSigningError):
    sc_sign_rsa('test', WRONG_MECH, VALID_KEY_ID, VALID_PIN, pkcs11=pkcs11)


def test_sc_sign_rsa_pkcs_pss_sha256_string_data(pkcs11):
  signature = sc_sign_rsa_pkcs_pss_sha256('test', VALID_KEY_ID, VALID_PIN, pkcs11=pkcs11)
  assert signature
  assert isinstance(signature, bytes)


def test_sc_sign_rsa_pkcs_pss_sha256_bytes_data(pkcs11):
  signature = sc_sign_rsa_pkcs_pss_sha256(b'test', VALID_KEY_ID, VALID_PIN, pkcs11=pkcs11)
  assert signature
  assert isinstance(signature, bytes)



from PyKCS11 import (CKG_MGF1_SHA256, CKM_SHA256, CKM_SHA256_RSA_PKCS_PSS,
                     CKM_SHA512, RSA_PSS_Mechanism)

MOCK_PYKCS11 = True  # Set to False to test it with real PyKCS11Lib and smart card

VALID_KEY_ID = (0x01,)
WRONG_KEY_ID = (0x20,)

VALID_PIN = '123456'
WRONG_PIN = 'xxxxxx'

VALID_MECH = RSA_PSS_Mechanism(CKM_SHA256_RSA_PKCS_PSS, CKM_SHA256, CKG_MGF1_SHA256, 32)
WRONG_MECH = RSA_PSS_Mechanism(CKM_SHA256_RSA_PKCS_PSS, CKM_SHA512, CKG_MGF1_SHA256, 32)



from setuptools import find_packages, setup

PACKAGE_NAME = 'oll-sc'
VERSION = '0.1.0'
AUTHOR = 'Open Law Library'
AUTHOR_EMAIL = 'info@openlawlib.org'
DESCRIPTION = 'Wrapper around PyKCS11 and openSC used to create smart card signatures'
URL = 'https://github.com/platform/oll-sc/tree/master'

with open('README.md', encoding='utf-8') as file_object:
  long_description = file_object.read()

packages = find_packages()

ci_require = [
    "pylint==2.3.1",
    "bandit==1.6.0",
    "pytest==4.5.0",
    "pytest-cov==2.7.1"
]

dev_require = [
    "autopep8==1.4.4",
    "pylint==2.3.1",
    "bandit==1.6.0"
]

tests_require = [
    "pytest==4.5.0"
]

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    packages=packages,
    include_package_data=True,
    data_files=[
        ('lib/site-packages/oll_sc', [
            './LICENSE.txt',
            './README.md'
        ])
    ],
    zip_safe=False,
    install_requires=[
        'click==6.7',
        'cryptography==2.3.1',
        'PyKCS11==1.5.5'
    ],
    extras_require={
        'ci': ci_require,
        'dev': dev_require,
        'test': tests_require
    },
    tests_require=tests_require,
    entry_points={
        'console_scripts': [
            'oll-sc = oll_sc.cli:oll_sc'
        ]
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Security',
        'Topic :: Software Development',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ]
)



import logging
import os
import platform
from functools import wraps
from pathlib import Path

from PyKCS11 import PyKCS11Lib

from .exceptions import PlatformNotSupported

logger = logging.getLogger(__name__)

# prebuild opensc libraries
OPENSC_LIBS_PATHS = {
    'Darwin-64bit': 'opensc/opensc-pkcs11-Darwin-64bit.so',
    'Windows-64bit': 'opensc/opensc-pkcs11-Windows-64bit.dll',
    'Windows-32bit': 'opensc/opensc-pkcs11-Windows-32bit.dll'
}

# https://github.com/easybuilders/easybuild/wiki/OS_flavor_name_version
# One of ['Windows', 'Darwin', 'Linux']
_SYSTEM = platform.system()
# One of ['32bit' , '64bit'] - depends on python executable
_ARCHITECTURE = platform.architecture()[0]

PLATFORM = '{}-{}'.format(_SYSTEM, _ARCHITECTURE)

logger.info('Platform: %s', PLATFORM)

# opensc-pkcs11 lib absolute path
OPENSC_LIB_PATH = Path(__file__).parent / OPENSC_LIBS_PATHS.get(PLATFORM, '')


def init_pkcs11(api_func):
  """Decorator to reinitialize PyKCS11 lib. Needed for long running processes.
  """
  @wraps(api_func)
  def decorator(*args, **kwargs):
    """If pkcs11 is NOT passed in kwargs, instantiate it and add it to
    kwargs.
                  NOTE: pkcs11 MUST be passed as kwarg!
    """
    pkcs11 = kwargs.pop('pkcs11', None)
    if pkcs11 is None:
      if not OPENSC_LIB_PATH.is_file():
        raise PlatformNotSupported(
            'opensc-pkcs11 library for platform {} is not included'
            .format(PLATFORM))

      pkcs11 = PyKCS11Lib()
      pkcs11.load(str(OPENSC_LIB_PATH.resolve()))
      logger.debug('PyKCS11Lib successfully loaded OpenSC library.')
    kwargs['pkcs11'] = pkcs11

    return api_func(*args, **kwargs)
  return decorator



import logging
from contextlib import contextmanager

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from PyKCS11 import (CKA_CERTIFICATE_TYPE, CKA_CLASS, CKA_ID, CKA_VALUE,
                     CKC_X_509, CKF_RW_SESSION, CKF_SERIAL_SESSION,
                     CKG_MGF1_SHA256, CKM_SHA256, CKM_SHA256_RSA_PKCS_PSS,
                     CKO_CERTIFICATE, CKO_PRIVATE_KEY, CKO_PUBLIC_KEY,
                     PyKCS11Error, RSA_PSS_Mechanism)

from . import init_pkcs11
from .exceptions import (SmartCardFindKeyObjectError, SmartCardNotPresentError,
                         SmartCardSigningError, SmartCardWrongPinError)

logger = logging.getLogger(__name__)


@init_pkcs11
def sc_export_pub_key_pem(key_id, pin, pkcs11=None):
  """Export public key for provided key id from smart card.

  Args:
    - key_id(tuple): Key ID as tuple (e.g. (1,))
    - pin(str): Pin for session login
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    Public key in PEM format (bytes)

  Raises:
    - SmartCardNotPresentError: If smart card is not inserted
    - SmartCardWrongPinError: If pin is incorrect
    - SmartCardFindKeyObjectError: If public key for given key id does not exist
  """
  with sc_session(pin, pkcs11=pkcs11) as session:
    try:
      pub_key = session.findObjects([(CKA_ID, key_id), (CKA_CLASS, CKO_PUBLIC_KEY)])[0]
      pub_key_value = session.getAttributeValue(pub_key, [CKA_VALUE])[0]

      pub_key_der = serialization.load_der_public_key(bytes(pub_key_value), default_backend())
      # Convert public key DER to PEM format
      pub_key_pem = pub_key_der.public_bytes(
          serialization.Encoding.PEM,
          serialization.PublicFormat.SubjectPublicKeyInfo,
      )

      logger.debug('Public key for key id: %s is \n%s', key_id, pub_key_pem.decode())
      return pub_key_pem
    except (IndexError, TypeError, ValueError):
      raise SmartCardFindKeyObjectError(key_id)


@init_pkcs11
def sc_export_x509_pem(key_id, pin, pkcs11=None):
  """Export x509 certificate for provided key id from smart card.

  Args:
    - key_id(tuple): Key ID as tuple (e.g. (1,))
    - pin(str): Pin for session login
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    x509 certificate in PEM format (bytes)

  Raises:
    - SmartCardNotPresentError: If smart card is not inserted
    - SmartCardWrongPinError: If pin is incorrect
    - SmartCardFindKeyObjectError: If x509 certificate for given key id does not exist
  """
  with sc_session(pin, pkcs11=pkcs11) as session:
    try:
      x509_cert = session.findObjects([(CKA_ID, key_id),
                                       (CKA_CLASS, CKO_CERTIFICATE),
                                       (CKA_CERTIFICATE_TYPE, CKC_X_509)])[0]
      x509_cert_value = session.getAttributeValue(x509_cert, [CKA_VALUE])[0]
      x509_cert_value_der = x509.load_der_x509_certificate(bytes(x509_cert_value),
                                                           default_backend())
      # Convert x509 certificate DER to PEM format
      x509_cert_value_pem = x509_cert_value_der.public_bytes(serialization.Encoding.PEM)

      logger.debug('X509 certificate for key id: %s is \n%s', key_id, x509_cert_value_pem.decode())
      return x509_cert_value_pem
    except (IndexError, TypeError, ValueError):
      raise SmartCardFindKeyObjectError(key_id)


@init_pkcs11
def sc_is_present(pkcs11=None):
  """Check if smart card is inserted.

  Args:
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    True if smart card is inserted otherwise False (bool)
  """
  return bool(pkcs11.getSlotList(tokenPresent=True))


@contextmanager
@init_pkcs11
def sc_session(pin, pkcs11=None):
  """Open token session needed for signing, encryption, etc.

  Args:
    - pin(str): Pin for session login
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    Session object (pykcs11.Session)

  Raises:
    - SmartCardNotPresentError: If smart card is not inserted
    - SmartCardWrongPinError: If pin is incorrect
  """
  if not sc_is_present(pkcs11=pkcs11):
    raise SmartCardNotPresentError('Please insert your smart card.')

  try:
    slot = pkcs11.getSlotList(tokenPresent=True)[0]

    session = pkcs11.openSession(slot, CKF_SERIAL_SESSION | CKF_RW_SESSION)
    logger.debug('Session opened for slot %s', slot)

    session.login(pin)
    yield session
    session.logout()

    logger.debug('Successfully logged out of session.')
  except PyKCS11Error:
    raise SmartCardWrongPinError('PIN is not valid.')
  finally:
    session.closeSession()
    logger.debug('Successfully closed the session.')


@init_pkcs11
def sc_sign_rsa(data, mechanism, key_id, pin, pkcs11=None):
  """Create and return signature using provided rsa mechanism.

  Args:
    - data(str | bytes): Data to be digested and signed
    - mechanism(PyKCS11 mechanism): Consult PyKCS11 for more info
    - key_id(tuple): Key ID as tuple (e.g. (1,))
    - pin(str): Pin for session login
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    Signature based on provided arguments (bytes)

  Raises:
    - SmartCardNotPresentError: If smart card is not inserted
    - SmartCardWrongPinError: If pin is incorrect
    - SmartCardFindKeyObjectError: If private key for given key id does not exist
    - SmartCardSigningError: If error happened during signing data
  """
  if isinstance(data, str):
    data = data.encode()

  logger.debug('About to sign data %s with mechanism %s', data, mechanism)

  with sc_session(pin, pkcs11=pkcs11) as session:
    try:
      priv_key = session.findObjects([(CKA_ID, key_id), (CKA_CLASS, CKO_PRIVATE_KEY)])[0]
      return session.sign(priv_key, data, mechanism)
    except (IndexError, TypeError):
      raise SmartCardFindKeyObjectError(key_id)
    except PyKCS11Error:
      raise SmartCardSigningError(data)


@init_pkcs11
def sc_sign_rsa_pkcs_pss_sha256(data, key_id, pin, pkcs11=None):
  """Sign data using SHA256_RSA_PKCS_PSS mechanism.

  Args:
    - data(str | bytes): Data to be digested and signed
    - key_id(tuple): Key ID as tuple (e.g. (1,))
    - pin(str): Pin for session login
    - pkcs11(PyKCS11): Automatically initialied; do not pass this argument

  Returns:
    Signature based on RSASSA-PSS signing algorithm on SHA256 digested data (bytes)

  Raises:
    - SmartCardNotPresentError: If smart card is not inserted
    - SmartCardWrongPinError: If pin is incorrect
    - SmartCardFindKeyObjectError: If private key for given key id does not exist
    - SmartCardSigningError: If error happened during signing data
  """
  mechanism = RSA_PSS_Mechanism(CKM_SHA256_RSA_PKCS_PSS, CKM_SHA256, CKG_MGF1_SHA256, 32)
  return bytes(sc_sign_rsa(data, mechanism, key_id, pin, pkcs11=pkcs11))
